/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Common.h"

DMACh *gif;

void gsInit() {
	gif = (DMACh*)&psH[0xA000];
}

void gsReset() {
	CSRr = 0xc;
	IMR = 0xff00;
}

void CSRwrite(u32 value) {
//	SysPrintf("GS_CSR: %x (CSRr=%x)\n", value, CSRr);
	CSRw = value;
	if (value & 0x200) { // resetGS
		CSRr = 0xc;
		IMR = 0xff00;
	} else {
		if (value & 0x8) CSRr&= ~0x8;
		if (value & 0x4) CSRr&= ~0x4;
	}
}

void IMRwrite(u32 value) {
//	SysPrintf("GS_IMR: %x\n", value);
	IMR = value;
	if (IMR & 0x8000) IMR = 0;
}

void gsWrite8(u32 mem, u8 value) {
	GSwrite8(mem, value);

	switch (mem) {
		case 0x12001000: // GS_CSR
			CSRwrite((CSRw & ~0x000000ff) | value); break;
		case 0x12001001: // GS_CSR
			CSRwrite((CSRw & ~0x0000ff00) | (value <<  8)); break;
		case 0x12001002: // GS_CSR
			CSRwrite((CSRw & ~0x00ff0000) | (value << 16)); break;
		case 0x12001003: // GS_CSR
			CSRwrite((CSRw & ~0xff000000) | (value << 24)); break;

/*		case 0x12001010: // GS_IMR
			IMRwrite(value);
			break;*/
	}

#ifdef HW_LOG
	HW_LOG("GS write 8 at %8.8lx with data %8.8lx\n", mem, value);
#endif
}

void gsWrite16(u32 mem, u16 value) {
	GSwrite16(mem, value);

/*	switch (mem) {
		case 0x12001000: // GS_CSR
			CSRwrite(value);
			break;

		case 0x12001010: // GS_IMR
			IMRwrite(value);
			break;
	}*/

#ifdef HW_LOG
	HW_LOG("GS write 16 at %8.8lx with data %8.8lx\n", mem, value);
#endif
}

void gsWrite32(u32 mem, u32 value) {
	GSwrite32(mem, value);

	switch (mem) {
		case 0x12001000: // GS_CSR
			CSRwrite(value);
			break;

		case 0x12001010: // GS_IMR
			IMRwrite(value);
			break;
	}

#ifdef HW_LOG
	HW_LOG("GS write 32 at %8.8lx with data %8.8lx\n", mem, value);
#endif
}

void gsWrite64(u32 mem, u64 value) {
	GSwrite64(mem, value);

	switch (mem) {
		case 0x12001000: // GS_CSR
			CSRwrite((u32)value);
			break;

		case 0x12001010: // GS_IMR
			IMRwrite((u32)value);
			break;
	}

#ifdef HW_LOG
	HW_LOG("GS write 64 at %8.8lx with data %8.8lx_%8.8lx\n", mem, ((u32*)&value)[1], (u32)value);
#endif
}

u8 gsRead8(u32 mem) {
	u8 ret=0;

	switch (mem) {
		case 0x12001000: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread8(mem);
			ret|= (u8)CSRr; break;
		case 0x12001001: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread8(mem);
			ret|= (u8)(CSRr >>  8); break;
		case 0x12001002: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread8(mem);
			ret|= (u8)(CSRr >> 16); break;
		case 0x12001003: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread8(mem);
			ret|= (u8)(CSRr >> 24); break;
		default:
			ret = GSread8(mem);
			break;
	}

#ifdef HW_LOG
//	HW_LOG("GS read 8 %8.8lx, at %8.8lx\n", ret, mem);
#endif
	return ret;
}

u16 gsRead16(u32 mem) {
	u16 ret=0;

	switch (mem) {
		case 0x12001000: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread16(mem);
			ret|= (u16)CSRr; break;
		case 0x12001002: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread16(mem);
			ret|= (u16)(CSRr >> 16); break;

		default:
			ret = GSread16(mem);
			break;
	}

#ifdef HW_LOG
//	HW_LOG("GS read 16 %8.8lx, at %8.8lx\n", ret, mem);
#endif
	return ret;
}

u32 gsRead32(u32 mem) {
	u32 ret=0;

	switch (mem) {
		case 0x12001000: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread32(mem);
			ret|= CSRr;
			if (Config.InterlaceHack) {
				CSRr^= 1<<13;
			}
			break;

		default:
			ret = GSread32(mem);
			break;
	}

#ifdef HW_LOG
//	HW_LOG("GS read 32 %8.8lx, at %8.8lx\n", ret, mem);
#endif
	return ret;
}

u64 gsRead64(u32 mem) {
	u64 ret=0;

	switch (mem) {
		case 0x12001000: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread64(mem);
			ret|= CSRr;
			if (Config.InterlaceHack) {
				CSRr^= 1<<13;
			}
			break;

		default:
			ret = GSread64(mem);
			break;
	}

#ifdef HW_LOG
//	HW_LOG("GS read 64 %8.8lx, at %8.8lx\n", ret, mem);
#endif
	return ret;
}

void gsIrq() {
	hwIntcIrq(0);
}

int  gsInterrupt() {
#ifdef GS_LOG 
	GS_LOG("gsInterrupt: %8.8x\n", cpuRegs.cycle);
#endif

	gif->chcr &=~ 0x100;
	hwDmacIrq(DMAC_GIF);
	psHu32(GIF_STAT)&= ~0xE00; // OPH=0 | APATH=0
	psHu32(GIF_STAT)&= ~0x1F000000; // QFC=0

	return 1;
}

int  _GIFchain() {
	u32 qwc = gif->qwc;
	u32 *pMem;

	if (qwc == 0) return 0;
    if(gif->madr & 0xf) gif->madr &= ~0xf;
	pMem = dmaGetAddr(gif->madr);
	if (pMem == NULL) return -1;

	GSgifTransfer3(pMem, qwc);
	gif->madr+= gif->qwc*16;
	gif->qwc = 0;
	return qwc;
}

#define GIFchain() \
	if (gif->qwc) { \
		_GIFchain(); \
		cycles+= (gif->qwc*4) / BIAS; /* guessing */ \
	}

void dmaGIF() {
	u32 *ptag;
	int done=0;
	int cycles=0;
	int id;

#ifdef GIF_LOG
	GIF_LOG("dmaGIF chcr = %lx, madr = %lx, qwc  = %lx\n"
			"        tadr = %lx, asr0 = %lx, asr1 = %lx\n",
			gif->chcr, gif->madr, gif->qwc,
			gif->tadr, gif->asr0, gif->asr1);
#endif

	if (psHu32(GIF_MODE) & 0x4) {
	} else
	if (vif1Regs->mskpath3 || psHu32(GIF_MODE) & 0x1) {
		gif->chcr &=~ 0x100;
		hwDmacIrq(2);
		return;
	}
	/*if ((psHu32(DMAC_CTRL) & 0xC0) == 0x80) { // STD == GIF
		gif->chcr &=~ 0x100;
		hwDmacIrq(2);
		return;
	}*/
	if ((psHu32(DMAC_CTRL) & 0xC) == 0xC/* &&
		(gif->tadr == psHu32(0xe050))*/) { // GIF MFIFO
		return;
	}

	psHu32(GIF_STAT)|= 0xE00; // OPH=1 | APATH=3
	psHu32(GIF_STAT)|= 0x1F000000; // FQC=31, hack ;)

	// Transfer Dn_QWC from Dn_MADR to GIF
	GIFchain();

	if ((gif->chcr & 0xc) == 0) { // Normal Mode
		INT(2, cycles);
		return;
	}

	// Chain Mode
	while (done == 0) {						 // Loop while Dn_CHCR.STR is 1
		ptag = (u32*)dmaGetAddr(gif->tadr);  //Set memory pointer to TADR
		if (ptag == NULL) {					 //Is ptag empty?
			psHu32(DMAC_STAT)|= 1<<15;		 //If yes, set BEIS (BUSERR) in DMAC_STAT register
			break;
		}
	
		// Transfer dma tag if tte is set
		if (gif->chcr & 0x40) {              
#ifdef GIF_LOG
			GIF_LOG("GIF TTE: %x_%x\n", ptag[3], ptag[2]);
#endif

//			Log=1;
//			GSgifTransfer3(ptag, 1); 
		}

		gif->chcr = ( gif->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );  //Transfer upper part of tag to CHCR bits 31-15

		id        = (ptag[0] >> 28) & 0x7;		//ID for DmaChain copied from bit 28 of the tag
		gif->qwc  = (u16)ptag[0];			    //QWC set to lower 16bits of the tag
		gif->madr = ptag[1];				    //MADR = ADDR field

#ifdef GIF_LOG
		GIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], gif->qwc, id, gif->madr);
#endif
		
		done = hwDmacSrcChainWithStack(gif, id);
		GIFchain();											 //Transfers the data set by the switch

		if (gif->chcr & 0x80 && ptag[0] >> 31) {			 //Check TIE bit of CHCR and IRQ bit of tag
#ifdef GIF_LOG
			GIF_LOG("dmaIrq Set\n");
#endif
//			SysPrintf("GSdmaIrq Set\n");
			break;
		}
	}
	INT(2, cycles);
	hwDmacIrq(2);
}

#undef GIFchain

static int cycles;

int mfifoGIFrbTransfer() {
	u32 maddr = psHu32(DMAC_RBOR);
	int msize = psHu32(DMAC_RBSR)+16;
	u32 *src;

	/* Check if the transfer should wrap around the ring buffer */
	if ((gif->madr+gif->qwc*16) > (maddr+msize)) {
		int s1 = (maddr+msize) - gif->madr;
		int s2 = gif->qwc*16 - s1;

		/* it does, so first copy 's1' bytes from 'addr' to 'data' */
		src = PSM(gif->madr);
		if (src == NULL) return -1;
		GSgifTransfer3(src, s1/16);

		/* and second copy 's2' bytes from 'maddr' to '&data[s1]' */
		src = PSM(maddr);
		if (src == NULL) return -1;
		GSgifTransfer3(src, s2/16);
	} else {
		/* it doesn't, so just transfer 'qwc*16' words 
		   from 'gif->madr' to GS */
		src = PSM(gif->madr);
		if (src == NULL) return -1;
		GSgifTransfer3(src, gif->qwc);
	}

	gif->madr+= gif->qwc*16;
	gif->madr = psHu32(DMAC_RBOR) + (gif->madr & psHu32(DMAC_RBSR));

	return 0;
}

int mfifoGIFchain() {
	u32 maddr = psHu32(DMAC_RBOR);
	int msize = psHu32(DMAC_RBSR)+16;
	u32 *pMem;

	/* Is QWC = 0? if so there is nothing to transfer */
	if (gif->qwc == 0) return 0;

	if (gif->madr >= maddr &&
		gif->madr < (maddr+msize)) {
		if (mfifoGIFrbTransfer() == -1) return -1;
	} else {
		pMem = (u32*)dmaGetAddr(gif->madr);
		if (pMem == NULL) return -1;
		GSgifTransfer3(pMem, gif->qwc);

		gif->madr+= gif->qwc*16;
	}

	cycles+= (gif->qwc*4) / BIAS; /* guessing */
	gif->qwc = 0;

	return 0;
}

void mfifoGIFtransfer(int qwc) {
	u32 *ptag;
	int id;
 
#ifdef GIF_LOG
	GIF_LOG("mfifoGIFtransfer qwc=0x%x\n", qwc);
#endif
 
	while (qwc > 0 && (gif->chcr & 0x100) == 0x100) {  // Loop while Dn_CHCR.STR is 1
		gif->tadr = psHu32(DMAC_RBOR) + (gif->tadr & psHu32(DMAC_RBSR));
		ptag = (u32*)dmaGetAddr(gif->tadr);
 
		id        = (ptag[0] >> 28) & 0x7;
		gif->qwc  = (u16)ptag[0];
		gif->madr = ptag[1];
 
#ifdef GIF_LOG
		GIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], gif->qwc, id, gif->madr);
#endif

//		if(gif->chcr & 0x40)
//			GSgifTransfer3(ptag, 1);

		switch (id) {
			case 0: // refe
				gif->tadr += 16;
				qwc = 0;
				gif->chcr &=~ 0x100;
				break;
 
			case 1: // cnt
				gif->madr = gif->tadr + 16;
				qwc-= gif->qwc + 1;
				// Set the taddr to the next tag
				gif->tadr += 16 + (gif->qwc * 16);
				break;
 
			case 3: // ref
			case 4: // refs
				gif->tadr += 16;
				qwc --;
				break;
 
			case 7: // end
				gif->madr = gif->tadr + 16;
				gif->tadr += 16;
				qwc = 0;
				gif->chcr &=~ 0x100;
				break;
		}

		if (mfifoGIFchain() == -1) {
			break;
		}

		if ((gif->chcr & 0x80) && (ptag[0] >> 31) && qwc > 0) {
#ifdef GIF_LOG
			GIF_LOG("dmaIrq Set\n");
#endif
			SysPrintf("mfifoGIFTransfer: dmaIrq Set\n");
		}
	}

	gif->tadr = psHu32(DMAC_RBOR) + (gif->tadr & psHu32(DMAC_RBSR));
	hwDmacIrq(2);
//	INT(2,cycles);
}

int gsFreeze(gzFile f, int Mode) {
	gzfreeze(&CSRw, sizeof(CSRw));
	gzfreeze(&CSRr, sizeof(CSRr));
	gzfreeze(&IMR, sizeof(IMR));

	return 0;
}

