/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "PsxCommon.h"


static int cnts = 6;

static void psxRcntUpd16(u32 index) {
	psxCounters[index].sCycle  = psxRegs.cycle;
	psxCounters[index].sCycleT = psxRegs.cycle;

	psxCounters[index].Cycle  = (0xffff - (psxRcntRcount16(index)&0xffff)) * psxCounters[index].rate;
	psxCounters[index].CycleT = (psxCounters[index].target - psxRcntRcount16(index)) * psxCounters[index].rate;
}

static void psxRcntUpd32(u32 index) {
	psxCounters[index].sCycle  = psxRegs.cycle;
	psxCounters[index].sCycleT = psxRegs.cycle;

	psxCounters[index].Cycle  = (0xffffffff - (psxRcntRcount32(index)&0xffffffff)) * psxCounters[index].rate;
	psxCounters[index].CycleT = (psxCounters[index].target - psxRcntRcount32(index)) * psxCounters[index].rate;
}

static void psxRcntReset16(u32 index) {
	psxCounters[index].count = 0;

	psxCounters[index].mode&= ~0x08301400;
	psxRcntUpd16(index);
}

static void psxRcntReset32(u32 index) {
	psxCounters[index].count = 0;
	psxCounters[index].otarget = 0x10000;

	psxCounters[index].mode&= ~0x08301400;
	psxRcntUpd32(index);
}

static void psxRcntSet() {
	u32 c;
	int i;

	if (Config.SafeCnts) {
		psxNextCounter = 0;
		psxNextsCounter = 0;
		return;
	}

	psxNextCounter = 0xffffffff;
	psxNextsCounter = psxRegs.cycle;

	for (i=0; i<cnts; i++) {
		c = psxCounters[i].Cycle - (psxRegs.cycle - psxCounters[i].sCycle);
		if (c < psxNextCounter) {
			psxNextCounter = c;
		}
		c = psxCounters[i].CycleT - (psxRegs.cycle - psxCounters[i].sCycleT);
		if (c < psxNextCounter) {
			psxNextCounter = c;
		}
	}
//	SysPrintf("psxRcntSet: %x (cycle)\n", psxNextCounter, psxRegs.cycle);
}

void psxRcntInit() {
	int i;

	memset(psxCounters, 0, sizeof(psxCounters));

	for (i=0; i<3; i++) {
		psxCounters[i].rate = 1;
		psxCounters[i].target = 0xffff;
	}
	for (i=3; i<6; i++) {
		psxCounters[i].rate = 1;
		psxCounters[i].target = 0xffffffff;
	}

	psxCounters[0].interrupt = 0x10;
	psxCounters[1].interrupt = 0x20;
	psxCounters[2].interrupt = 0x40;

	psxCounters[3].interrupt = 0x04000;
	psxCounters[4].interrupt = 0x08000;
	psxCounters[5].interrupt = 0x10000;

	if (SPU2async != NULL) {
		cnts = 7;

		psxCounters[6].rate = 768 * 64;
		psxCounters[6].target = 1;
		psxCounters[6].mode = 0x8;
	} else cnts = 6;

	for (i=0; i<3; i++)
		psxRcntUpd16(i);
	for (i=3; i<6; i++)
		psxRcntUpd32(i);
	for (i=6; i<cnts; i++)
		psxRcntUpd16(i);
	psxRcntSet();
}

void psxVSyncStart() {
	psxHu32(0x1070)|= 1;
}

void psxVSyncEnd() {
	cdvdVsync();
	psxHu32(0x1070)|= 0x800;
}

void _testRcnt16target(int i) {
	psxCounters[i].mode|= 0x0800; // Target flag
	if (psxCounters[i].mode & 0x10) { // Target interrupt
	psxCounters[i].mode&= ~0x1000;
#ifdef PSXCNT_LOG
		PSXCNT_LOG("[%d] target 0x%x >= 0x%x (CycleT); count=0x%x, target=0x%x\n", i, (psxRegs.cycle - psxCounters[i].sCycleT), psxCounters[i].CycleT, psxRcntRcount16(i), psxCounters[i].target);
#endif
		if (psxCounters[i].mode & 0x40) { // Repeat interrupt
			psxHu32(0x1070)|= psxCounters[i].interrupt;
			psxCounters[i].mode|= 0x0400; // Interrupt flag
		} else {
			if ((psxCounters[i].mode & 0x00100000) == 0) {
				psxCounters[i].mode|= 0x00100000; // Internal target interrupt flag
				psxHu32(0x1070)|= psxCounters[i].interrupt;
				psxCounters[i].mode|= 0x0400; // Interrupt flag
			}
		}
	}

	if (psxCounters[i].mode & 0x08) {
		psxRcntReset16(i);
	} else { // Reset on target
		psxCounters[i].mode|= 0x08000000;
	}
}

void _testRcnt16overflow(int i) {
	psxCounters[i].mode|= 0x1000; // Overflow flag
	if (psxCounters[i].mode & 0x0020) { // Overflow interrupt
#ifdef PSXCNT_LOG
		PSXCNT_LOG("[%d] overflow 0x%x >= 0x%x (Cycle); Rcount=0x%x, count=0x%x\n", i, (psxRegs.cycle - psxCounters[i].sCycle), psxCounters[i].Cycle, psxRcntRcount16(i), psxCounters[i].count);
#endif
		if (psxCounters[i].mode & 0x0040) { // Repeat interrupt
			psxHu32(0x1070)|= psxCounters[i].interrupt;
			psxCounters[i].mode|= 0x0400; // Interrupt flag
		} else {
			if ((psxCounters[i].mode & 0x200000) == 0) {
				psxCounters[i].mode|= 0x200000; // Internal Overflow interrupt flag
				psxHu32(0x1070)|= psxCounters[i].interrupt;
				psxCounters[i].mode|= 0x0400; // Interrupt flag
			}
		}
	}

	psxRcntReset16(i);
}

void _testRcnt32target(int i) {
	psxCounters[i].mode|= 0x0800; // Target flag
	psxCounters[i].mode&= ~0x1000;
	if (psxCounters[i].mode & 0x10) { // Target interrupt
#ifdef PSXCNT_LOG
		PSXCNT_LOG("[%d] target 0x%x >= 0x%x (CycleT); count=0x%x, target=0x%x\n", i, (psxRegs.cycle - psxCounters[i].sCycleT), psxCounters[i].CycleT, psxRcntRcount32(i), psxCounters[i].target);
#endif
		if (psxCounters[i].mode & 0x40) { // Repeat interrupt
			psxHu32(0x1070)|= psxCounters[i].interrupt;
			psxCounters[i].mode|= 0x0400; // Interrupt flag
		} else {
			if ((psxCounters[i].mode & 0x00100000) == 0) {
				psxCounters[i].mode|= 0x00100000; // Internal target interrupt flag
				psxHu32(0x1070)|= psxCounters[i].interrupt;
				psxCounters[i].mode|= 0x0400; // Interrupt flag
			}
		}
	}

	if (psxCounters[i].mode & 0x08) { // Reset on target
		psxRcntReset32(i);
	} else {
		psxCounters[i].mode|= 0x08000000;
	}
}

void _testRcnt32overflow(int i) {
	
		psxCounters[i].mode|= 0x1000; // Overflow flag
		if (psxCounters[i].mode & 0x0020) { // Overflow interrupt
#ifdef PSXCNT_LOG
			PSXCNT_LOG("[%d] overflow 0x%x >= 0x%x (Cycle); Rcount=0x%x, count=0x%x\n", i, (psxRegs.cycle - psxCounters[i].sCycle), psxCounters[i].Cycle, psxRcntRcount32(i), psxCounters[i].count);
#endif
			if (psxCounters[i].mode & 0x0040) { // Repeat interrupt
				psxHu32(0x1070)|= psxCounters[i].interrupt;
				psxCounters[i].mode|= 0x0400; // Interrupt flag
			} else {
				if ((psxCounters[i].mode & 0x200000) == 0) {
					psxCounters[i].mode|= 0x200000; // Internal Overflow interrupt flag
					psxHu32(0x1070)|= psxCounters[i].interrupt;
					psxCounters[i].mode|= 0x0400; // Interrupt flag
				}
			}
		}
		psxRcntReset32(i);
}

void _testRcnt16(int i) {
	if (Config.SafeCnts) {
		if (psxRcntRcount32(i) == psxCounters[i].target)
			_testRcnt16overflow(i);
	} else {
		if ((psxRegs.cycle - psxCounters[i].sCycle) >= psxCounters[i].Cycle)
			_testRcnt16overflow(i);
	}

	/* mode & 0x08000000 means target interrupt already happened, 
	   if so don't retest the target */
	if (psxCounters[i].mode & 0x08000000) return;

	if (Config.SafeCnts) {
		if (psxRcntRcount16(i) > psxCounters[i].target)
			_testRcnt16target(i);
	} else {
		if ((psxRegs.cycle - psxCounters[i].sCycleT) >= psxCounters[i].CycleT)
			_testRcnt16target(i);
	}
}

void _testRcnt32(int i) {
	if (Config.SafeCnts) {
		if (psxRcntRcount32(i) == psxCounters[i].target)
			_testRcnt32overflow(i);
	} else {
		if ((psxRegs.cycle - psxCounters[i].sCycleT) >= psxCounters[i].Cycle)
			_testRcnt32overflow(i);
	}

	/* mode & 0x08000000 means target interrupt already happened, 
	   if so don't retest the target */
	if (psxCounters[i].mode & 0x08000000) return;

	if (Config.SafeCnts) {
		if (psxRcntRcount32(i) > psxCounters[i].target)
			_testRcnt32target(i);
	} else {
		if ((psxRegs.cycle - psxCounters[i].sCycleT) >= psxCounters[i].CycleT)
			_testRcnt32target(i);
	}
}

void psxRcntUpdate() {
		_testRcnt16(0);
		_testRcnt16(1);
		_testRcnt16(2);
		_testRcnt32(3);
		_testRcnt32(4);
		_testRcnt32(5);

	if (cnts >= 7) {
		if ((psxRegs.cycle - psxCounters[6].sCycleT) >= psxCounters[6].CycleT) {
			SPU2async(psxRegs.cycle - psxCounters[6].sCycle);
			psxRcntReset16(6);
		}
	}

	psxRcntSet();
}

void psxRcntWcount16(int index, u32 value) {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("writeCcount[%d] = %x\n", index, value);
#endif
	psxCounters[index].count = value;
	psxRcntUpd16(index);
	psxRcntSet();
}

void psxRcntWcount32(int index, u32 value) {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("writeCcount[%d] = %x\n", index, value);
#endif
	psxCounters[index].count = value;
	psxRcntUpd32(index);
	psxRcntSet();
}

void psxRcnt0Wmode(u32 value)  {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("IOP writeCmode[0] = %lx\n", value);
#endif
	psxCounters[0].mode = value;
	
	
	psxCounters[0].rate = 1;
	

	if(value & 0x100) psxCounters[0].rate = PSXPIXEL;
	// Need to set a rate and target
	if(psxCounters[0].mode & 0x1){
		if(((psxCounters[0].mode & 0x6) >> 1) == 1) {
		psxRcntReset16(0);
		psxCounters[0].mode &= ~0x8101C00;
		psxRcntSet();
		}
		if(((psxCounters[0].mode & 0x6) >> 1) == 2) {
		psxRcntReset16(0);
		psxCounters[0].mode &= ~0x8101C00;
		}
	} else {
	psxCounters[0].mode &= ~0x8101C00;
	psxRcntUpd16(0);
	psxRcntSet();
	}
}

void psxRcnt1Wmode(u32 value)  {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("IOP writeCmode[1] = %lx\n", value);
#endif
	psxCounters[1].mode = value;

	psxCounters[1].rate = 1;
	
	if(value & 0x100)psxCounters[1].rate = PSXHBLANK;

	if(psxCounters[1].mode & 0x1){
		if(((psxCounters[1].mode & 0x6) >> 1) == 1) {
		psxRcntReset16(1);
		psxCounters[1].mode &= ~0x8101C00;
		psxRcntSet();
		}
		if(((psxCounters[1].mode & 0x6) >> 1) == 2) {
		psxRcntReset16(1);
		psxCounters[1].mode &= ~0x8101C00;
		}
	} else {
	psxCounters[1].mode &= ~0x8101C00;
	psxRcntUpd16(1);
	psxRcntSet();
	}
}

void psxRcnt2Wmode(u32 value)  {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("IOP writeCmode[2] = %lx\n", value);
#endif
	psxCounters[2].mode = value;

	switch(value & 0x200){
		case 0x200:
			psxCounters[2].rate = 8;
			break;
		case 0x000:
			psxCounters[2].rate = 1;
			break;
	}
	
	
	// Need to set a rate and target
	psxCounters[2].mode &= ~0x8101C00;
	psxRcntReset16(2);
	psxRcntSet();
}

void psxRcnt3Wmode(u32 value)  {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("IOP writeCmode[3] = %lx\n", value);
#endif
	psxCounters[3].mode = value;

	psxCounters[3].rate = 1;
	
	if(value & 0x100)psxCounters[3].rate = PSXHBLANK;

	if(psxCounters[3].mode & 0x1){
		if(((psxCounters[3].mode & 0x6) >> 1) == 1) {
		psxRcntReset32(3);
		psxCounters[3].mode &= ~0x8101C00;
		psxRcntSet();
		}
		if(((psxCounters[3].mode & 0x6) >> 1) == 2) {
		psxCounters[3].mode &= ~0x8101C00;
		psxRcntReset32(3);
		}
	} else {
	psxCounters[3].mode &= ~0x8101C00;
	psxRcntUpd32(3);
	psxRcntSet();
	}
}

void psxRcnt4Wmode(u32 value)  {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("IOP writeCmode[4] = %lx\n", value);
#endif
	psxCounters[4].mode = value;
	switch(value & 0x6000){
		case 0x0000:
            psxCounters[4].rate = 1;
			break;
		case 0x2000:
			psxCounters[4].rate = 8;
			break;
		case 0x4000:
			psxCounters[4].rate = 16;
			break;
		case 0x6000:
			psxCounters[4].rate = 256;
			break;
	}

	// Need to set a rate and target
	psxCounters[4].mode &= ~0x8101C00;
	psxRcntReset32(4);
	psxRcntSet();
}

void psxRcnt5Wmode(u32 value)  {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("IOP writeCmode[5] = %lx\n", value);
#endif
	psxCounters[5].mode = value;

	switch(value & 0x6000){
		case 0x0000:
            psxCounters[5].rate = 1;
			break;
		case 0x2000:
			psxCounters[5].rate = 8;
			break;
		case 0x4000:
			psxCounters[5].rate = 16;
			break;
		case 0x6000:
			psxCounters[5].rate = 256;
			break;
	}

	// Need to set a rate and target
	psxCounters[5].mode &= ~0x8101C00;
	psxRcntReset32(5);
	psxRcntSet();
}

void psxRcntWtarget16(int index, u32 value) {
#ifdef PSXCNT_LOG
	PSXCNT_LOG("writeCtarget[%ld] = %lx\n", index, value);
#endif
	psxCounters[index].target = value;
	psxCounters[index].mode&= ~0x08000000;
	//psxCounters[index].sCycleT = psxRegs.cycle;
	psxCounters[index].CycleT = ((psxCounters[index].target - psxRcntRcount16(index)) * psxCounters[index].rate);
//	psxRcntUpd16(index);
	psxRcntSet();
}

void psxRcntWtarget32(int index, u32 value) {
	psxCounters[index].target = value;
	psxCounters[index].mode&= ~0x08000000;
	//psxCounters[index].sCycleT = psxRegs.cycle;
	psxCounters[index].CycleT = ((psxCounters[index].target - psxRcntRcount32(index)) * psxCounters[index].rate);
#ifdef PSXCNT_LOG
	PSXCNT_LOG("writeCtarget32[%ld] = %lx (count=%lx) ; sCycleT: %x CycleT: %x\n",
			   index, value, psxRcntRcount32(index), psxCounters[index].sCycleT, psxCounters[index].CycleT);
#endif
//	psxRcntUpd32(index);
	psxRcntSet();
}

u16 psxRcntRcount16(int index) {
	return (u16)(psxCounters[index].count + ((psxRegs.cycle - psxCounters[index].sCycle) / psxCounters[index].rate));
}

u32 psxRcntRcount32(int index) {
	return psxCounters[index].count + ((psxRegs.cycle - psxCounters[index].sCycle) / psxCounters[index].rate);
}

int psxRcntFreeze(gzFile f, int Mode) {
	gzfreezel(psxCounters);

	return 0;
}
