/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "Common.h"
#include "SPR.h"

DMACh *spr0;
DMACh *spr1;

void sprInit() {
	spr0 = (DMACh*)&psH[0xD000];
	spr1 = (DMACh*)&psH[0xD400];
}

__inline static void SPR0transfer(u32 *data, int size) {
/*	while (size > 0) {
#ifdef SPR_LOG
		SPR_LOG("SPR1transfer: %x\n", *data);
#endif
		data++; size--;
	}*/
	if ((psHu32(DMAC_CTRL) & 0xC) == 0xC || // GIF MFIFO
		(psHu32(DMAC_CTRL) & 0xC) == 0x8) { // VIF1 MFIFO
		hwMFIFOWrite(spr0->madr, (u8*)&psS[spr0->sadr & 0x3fff], size*4);
	} else {
		memcpy(data, &psS[spr0->sadr & 0x3fff], size*4);
	}
	spr0->sadr+= size*4;
}

int  _SPR0chain() {
	u32 qwc = spr0->qwc;
	u32 *pMem;

	if (qwc == 0) return 0;

	pMem = dmaGetAddr(spr0->madr);
	if (pMem == NULL) return -1;

	SPR0transfer(pMem, qwc*4);
	spr0->madr+= spr0->qwc*16;
	spr0->qwc = 0;
	return 0;
}

#define SPR0chain() \
	if (spr0->qwc) { \
		_SPR0chain(); \
/*		cycles+= dma->qwc / BIAS;*/ /* guessing */ \
	}

void _SPR0interleave() {
	int qwc = spr0->qwc;
	int sqwc = psHu32(DMAC_SQWC) & 0xff;
	int tqwc = (psHu32(DMAC_SQWC) >> 16) & 0xff;

	while (qwc > 0) {
		spr0->qwc = tqwc; qwc-= tqwc;
		_SPR0chain();

		spr0->madr+= sqwc*16; qwc-= sqwc;
	}

	spr0->chcr &=~ 0x100;
	hwDmacIrq(8);
}

void _dmaSPR0() {
	u32 *ptag;
	int id;

	if ((spr0->chcr & 0xc) == 0x8) { // Interleave Mode
		_SPR0interleave();
		return;
	}

	// Transfer Dn_QWC from SPR to Dn_MADR
	SPR0chain();
	

	if ((spr0->chcr & 0xc) == 0) { // Normal Mode
		spr0->chcr&= ~0x100;
		hwDmacIrq(8);
		return;
	}

	// Destination Chain Mode

	while (spr0->chcr & 0x100) {  // Loop while Dn_CHCR.STR is 1
		ptag = (u32*)&psS[spr0->sadr & 0x3fff];
		spr0->sadr+= 16;

		// Transfer dma tag if tte is set
//		if (spr0->chcr & 0x40) SPR0transfer(ptag, 4);

		spr0->chcr = ( spr0->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );	//Transfer upper part of tag to CHCR bits 31-15

		id        = (ptag[0] >> 28) & 0x7;		//ID for DmaChain copied from bit 28 of the tag
		spr0->qwc  = (u16)ptag[0];				//QWC set to lower 16bits of the tag
		spr0->madr = ptag[1];					//MADR = ADDR field

#ifdef SPR_LOG
		SPR_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], spr0->qwc, id, spr0->madr);
#endif
		
		switch (id) {
			case 0: // CNTS - Transfer QWC following the tag (Stall Control)
				psHu32(DMAC_STADR) = spr0->madr;					//Copy MADR to DMAC_STADR stall addr register
				break;

			case 1: // CNT - Transfer QWC following the tag.
				break;

			case 7: // End - Transfer QWC following the tag
				spr0->chcr &=~ (1<<8);								//End Transfer
				break;
		}
		SPR0chain();

/*		if (spr0->chcr & 0x80 && ptag[0] >> 31) {
#ifdef SPR_LOG
			SPR_LOG("dmaIrq Set\n");
#endif
			spr0->chcr&= ~0x100;
			hwDmacIrq(8);
			return;
		}*/
	}

	spr0->chcr &=~ 0x100;
	hwDmacIrq(8);
}

void dmaSPR0() { // fromSPR
	int qwc = spr0->qwc;
#ifdef SPR_LOG
	SPR_LOG("dmaSPR0 chcr = %lx, madr = %lx, qwc  = %lx, sadr = %lx\n",
			spr0->chcr, spr0->madr, spr0->qwc, spr0->sadr);
#endif

	_dmaSPR0();
	if ((psHu32(DMAC_CTRL) & 0xC) == 0xC) { // GIF MFIFO
		spr0->madr = psHu32(DMAC_RBOR) + (spr0->madr & (psHu32(DMAC_RBSR)+15));
		mfifoGIFtransfer(qwc);
	} else
	if ((psHu32(DMAC_CTRL) & 0xC) == 0x8) { // VIF1 MFIFO
		spr0->madr = psHu32(DMAC_RBOR) + (spr0->madr & (psHu32(DMAC_RBSR)+15));
		mfifoVIF1transfer(qwc);
	}
}

__inline static void SPR1transfer(u32 *data, int size) {
/*	{
		int i;
		for (i=0; i<size; i++) {
#ifdef SPR_LOG 
		   SPR_LOG( "SPR1transfer[0x%x]: 0x%x\n", (spr1->sadr+i*4) & 0x3fff, data[i] );
#endif
		}
	}*/
	memcpy(&psS[spr1->sadr & 0x3fff], data, size*4);
	spr1->sadr+= size*4;
}

int  _SPR1chain() {
	u32 qwc = spr1->qwc;
	u32 *pMem;

	if (qwc == 0) return 0;

	pMem = dmaGetAddr(spr1->madr);
	if (pMem == NULL) return -1;

	SPR1transfer(pMem, qwc*4);
	spr1->madr+= spr1->qwc*16;
	spr1->qwc = 0;
	return qwc;
}

#define SPR1chain() \
	if (spr1->qwc) { \
		_SPR1chain(); \
/*		cycles+= dma->qwc / BIAS;*/ /* guessing */ \
	}

void _SPR1interleave() {
	int qwc = spr1->qwc;
	int sqwc = psHu32(DMAC_SQWC) & 0xff;
	int tqwc = (psHu32(DMAC_SQWC) >> 16) & 0xff;

	while (qwc > 0) {
		spr1->qwc = tqwc; qwc-= tqwc;
		SPR1chain();

		spr1->madr+= sqwc*16; qwc-= sqwc;
	}

	spr1->chcr &=~ 0x100;
	hwDmacIrq(9);
}

void dmaSPR1() { // toSPR
	u32 *ptag;
	int id, done=0;

#ifdef SPR_LOG
	SPR_LOG("dmaSPR1 chcr = 0x%x, madr = 0x%x, qwc  = 0x%x\n"
			"        tadr = 0x%x, sadr = 0x%x\n",
			spr1->chcr, spr1->madr, spr1->qwc,
			spr1->tadr, spr1->sadr);
#endif

	if ((spr1->chcr & 0xc) == 0x8) { // Interleave Mode
		_SPR1interleave();
		return;
	}

	// Transfer Dn_QWC from Dn_MADR to SPR1
	SPR1chain();

	if ((spr1->chcr & 0xc) == 0) { // Normal Mode
		spr1->chcr &=~ 0x100;
		hwDmacIrq(9);
		return;
	}

	// Chain Mode

	while (done == 0) {  // Loop while Dn_CHCR.STR is 1
		ptag = (u32*)dmaGetAddr(spr1->tadr);		//Set memory pointer to TADR
		if (ptag == NULL) {							//Is ptag empty?
			psHu32(DMAC_STAT)|= 1<<15;				//If yes, set BEIS (BUSERR) in DMAC_STAT register
			break;
		}

		// Transfer dma tag if tte is set
		if (spr1->chcr & 0x40) {
#ifdef SPR_LOG
			SPR_LOG("SPR TTE: %x_%x\n", ptag[3], ptag[2]);
#endif
			SPR1transfer(ptag, 4);				//Transfer Tag
		}

		spr1->chcr = ( spr1->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );	//Transfer upper part of tag to CHCR bits 31-15

		id        = (ptag[0] >> 28) & 0x7;			//ID for DmaChain copied from bit 28 of the tag
		spr1->qwc  = (u16)ptag[0];					//QWC set to lower 16bits of the tag
		spr1->madr = ptag[1];						//MADR = ADDR field

#ifdef SPR_LOG
		SPR_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], spr1->qwc, id, spr1->madr);
#endif
		
		done = hwDmacSrcChain(spr1, id);
		SPR1chain();										//Transfers the data set by the switch

		if (spr1->chcr & 0x80 && ptag[0] >> 31) {			//Check TIE bit of CHCR and IRQ bit of tag
#ifdef SPR_LOG
			SPR_LOG("dmaIrq Set\n");
#endif
			SysPrintf("SPR1dmaIrq Set\n");
			break;
		}
	}

	spr1->chcr &=~ 0x100;
	hwDmacIrq(9);
}

